from django.apps import apps as global_apps
from django.db import DEFAULT_DB_ALIAS


def create_permissions(app_config, verbosity=2, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs):
    if not app_config.models_module:
        return

    if app_config.label != "saas_base":
        return

    try:
        Permission = apps.get_model(app_config.label, "Permission")
    except LookupError:
        return

    existed_perms = set(Permission.objects.values_list("name", flat=True).all())

    perms = []
    if "tenant.read" not in existed_perms:
        perms.append(Permission(name="tenant.read", description="Read permission for tenants"))
    if "tenant.write" not in existed_perms:
        perms.append(Permission(name="tenant.write", description="Write permission for tenants"))
    if "tenant.admin" not in existed_perms:
        perms.append(Permission(name="tenant.admin", description="Admin permission for tenants"))

    if perms:
        Permission.objects.using(using).bulk_create(perms, ignore_conflicts=True)
    if verbosity >= 2:
        for perm in perms:
            print(f"Adding saas_base.Permission '{perm.name}'")



import uuid
import unicodedata
from typing import List, Set
from django.db import models
from django.conf import settings
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from django.utils.functional import cached_property
from .group import Group
from .permission import Permission
from ..db import CachedManager


class MemberManager(CachedManager):
    natural_key = ["tenant_id", "user_id"]

    def get_by_natural_key(self, tenant_id, user_id) -> "Member":
        return self.get_from_cache_by_natural_key(tenant_id, user_id)

    def create_owner(self, user, tenant) -> "Member":
        return self.create(
            user=user,
            name=user.get_full_name(),
            tenant=tenant,
            is_owner=True,
            status=Member.InviteStatus.ACTIVE,
        )


class Member(models.Model):
    username_validator = UnicodeUsernameValidator()

    class InviteStatus(models.IntegerChoices):
        REQUEST = 0, 'request'
        WAITING = 1, 'waiting'
        ACTIVE = 2, 'active'

    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)
    tenant = models.ForeignKey(settings.SAAS_TENANT_MODEL, on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, blank=True, null=True)
    name = models.CharField(max_length=150, blank=True)

    # this email is only used for invitation
    invite_email = models.EmailField(null=True, blank=True)
    inviter = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='+',
    )

    status = models.SmallIntegerField(default=InviteStatus.REQUEST, choices=InviteStatus.choices)
    is_owner = models.BooleanField(default=False, db_index=True)

    created_at = models.DateTimeField(default=timezone.now, db_index=True)

    groups = models.ManyToManyField(
        Group,
        blank=True,
        help_text=_(
            "The groups this user belongs to. A user will get all permissions "
            "granted to each of their groups."
        ),
    )
    permissions = models.ManyToManyField(
        Permission,
        blank=True,
        help_text=_("Specific permissions for this user."),
    )
    objects = MemberManager()

    class Meta:
        verbose_name = _("member")
        verbose_name_plural = _("members")
        unique_together = [
            ['tenant', 'user'],
        ]
        ordering = ['-created_at']
        db_table = 'saas_member'

    def __str__(self):
        return self.name or self.invite_email

    @property
    def is_active(self) -> bool:
        return self.status == self.InviteStatus.ACTIVE

    @classmethod
    def normalize_username(cls, username: str):
        return unicodedata.normalize("NFKC", username)

    @cached_property
    def group_permissions(self) -> List[str]:
        return self.__get_group_permissions()

    @cached_property
    def user_permissions(self) -> List[str]:
        return self.__get_user_permissions()

    def get_all_permissions(self) -> Set[str]:
        return set(self.user_permissions + self.group_permissions)

    def __get_group_permissions(self):
        values = self.groups.values_list('permissions__name')
        names = [v[0] for v in values]
        return names

    def __get_user_permissions(self):
        values = self.permissions.all().values_list('name')
        names = [v[0] for v in values]
        return names



from email.utils import formataddr
from django.utils.translation import gettext_lazy as _
from django.conf import settings
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.throttling import AnonRateThrottle
from ..drf.views import Endpoint
from ..settings import saas_settings
from ..models import UserEmail
from ..security import check_security_rules
from ..serializers.password import (
    PasswordForgetSerializer,
    PasswordResetSerializer,
)
from .._notification import send_mail


class PasswordForgotEndpoint(Endpoint):
    email_template_id = "reset_password"
    email_subject = _("Password Reset Request")

    permission_classes = []
    throttle_classes = [AnonRateThrottle]
    serializer_class = PasswordForgetSerializer

    def post(self, request: Request):
        """Send a forgot password reset email code."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj: UserEmail = serializer.save()

        # check bad request rules
        check_security_rules(saas_settings.RESET_PASSWORD_SECURITY_RULES, request)

        code = serializer.save_password_code(obj)
        name = obj.user.get_full_name() or obj.user.get_username()
        send_mail(
            self.__class__,
            self.email_subject,
            self.email_template_id,
            recipients=[formataddr((name, obj.email))],
            code=code,
            user=obj.user,
        )
        return Response(status=204)


class PasswordResetEndpoint(Endpoint):
    permission_classes = []
    throttle_classes = [AnonRateThrottle]
    serializer_class = PasswordResetSerializer

    def post(self, request: Request):
        """Reset password of a user with the given code."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response({"next": settings.LOGIN_URL})



import re
from django.core import mail
from django.test import override_settings
from saas_base.models import UserEmail, Member
from tests.client import FixturesTestCase


class TestSignUpWithoutCreateUser(FixturesTestCase):
    user_id = FixturesTestCase.ADMIN_USER_ID

    def get_auth_code(self):
        msg = mail.outbox[0]
        codes = re.findall(r'Code: (\w{6})', msg.body)
        return codes[0]

    def test_signup_success(self):
        data1 = {"username": "demo", "email": "hi@foo.com", "password": "hello world"}
        resp = self.client.post("/s/signup/request", data=data1)
        self.assertEqual(resp.status_code, 204)
        self.assertEqual(len(mail.outbox), 1)
        data2 = {**data1, "code": self.get_auth_code()}
        resp = self.client.post("/s/signup/confirm", data=data2)
        self.assertEqual(resp.status_code, 200)

    def test_signup_existed_email(self):
        user = self.get_user()
        UserEmail.objects.create(user=user, email="hi@foo.com", primary=True, verified=True)
        data = {"username": "foo", "email": "hi@foo.com", "password": "hello world"}
        resp = self.client.post("/s/signup/request", data=data)
        self.assertEqual(resp.status_code, 400)
        self.assertIn('existing', resp.json()['email'][0])

    def test_signup_blocked_email(self):
        rules = [
            {
                'backend': 'saas_base.security.rules.BlockedEmailDomains',
                'options': {
                    'domains': ['bar.com']
                }
            }
        ]
        with override_settings(SAAS={"SIGNUP_SECURITY_RULES": rules}):
            data = {"username": "bar", "email": "hi@bar.com", "password": "hello world"}
            resp = self.client.post("/s/signup/request", data=data)
            self.assertEqual(resp.status_code, 400)

    def test_signup_too_many_dots(self):
        rules = [{'backend': 'saas_base.security.rules.TooManyDots'}]
        with override_settings(SAAS={"SIGNUP_SECURITY_RULES": rules}):
            data = {"username": "bar", "email": "a.b.c.d.e.f@bar.com", "password": "hello world"}
            resp = self.client.post("/s/signup/request", data=data)
            self.assertEqual(resp.status_code, 400)

    def test_signup_turnstile(self):
        rules = [{'backend': 'saas_base.security.rules.Turnstile'}]
        with override_settings(SAAS={"SIGNUP_SECURITY_RULES": rules}):
            data = {"username": "bar", "email": "hi@bar.com", "password": "hello world"}
            resp = self.client.post("/s/signup/request", data=data)
            self.assertEqual(resp.status_code, 400)

            data = {**data, 'cf-turnstile-response': '**token**'}
            with self.mock_requests('turnstile_success.json'):
                resp = self.client.post("/s/signup/request", data=data)
                self.assertEqual(resp.status_code, 204)

            data = {**data, 'cf-turnstile-response': '**token**'}
            with self.mock_requests('turnstile_failed.json'):
                resp = self.client.post("/s/signup/request", data=data)
                self.assertEqual(resp.status_code, 400)

    def test_signup_with_membership_invite(self):
        # prepare membership
        email = "hi@foo.com"
        Member.objects.create(tenant_id=self.tenant_id, invite_email=email)

        data1 = {"username": "demo", "email": email, "password": "hello world"}
        self.client.post("/s/signup/request", data=data1)
        data2 = {**data1, "code": self.get_auth_code()}
        resp = self.client.post("/s/signup/confirm", data=data2)
        self.assertEqual(resp.status_code, 200)
        obj = UserEmail.objects.get(email=email)
        member = Member.objects.get(user_id=obj.user_id, tenant_id=self.tenant_id)
        self.assertEqual(member.status, Member.InviteStatus.WAITING)


@override_settings(SAAS={"SIGNUP_REQUEST_CREATE_USER": True})
class TestSignUpWithCreateUser(FixturesTestCase):
    def get_auth_code(self):
        msg = mail.outbox[0]
        codes = re.findall(r'Code: (\w{6})', msg.body)
        return codes[0]

    def test_signup_success(self):
        data1 = {"username": "demo", "email": "hi@foo.com", "password": "hello world"}
        resp = self.client.post("/s/signup/request", data=data1)
        self.assertEqual(resp.status_code, 204)
        obj = UserEmail.objects.get(email="hi@foo.com")
        self.assertEqual(obj.verified, False)

        data2 = {"code": self.get_auth_code(), "email": "hi@foo.com"}
        resp = self.client.post("/s/signup/confirm", data=data2)
        self.assertEqual(resp.status_code, 200)
        obj = UserEmail.objects.get(email="hi@foo.com")
        self.assertEqual(obj.verified, True)


class TestLoginAPI(FixturesTestCase):
    user_id = FixturesTestCase.OWNER_USER_ID

    def test_login_with_username(self):
        user = self.get_user()

        data = {"username": user.username, "password": "hello world"}
        resp = self.client.post("/s/login", data=data)
        self.assertEqual(resp.status_code, 400)

        user.set_password('hello world')
        user.save()

        resp = self.client.post("/s/login", data=data)
        self.assertEqual(resp.status_code, 200)
        self.assertIn("next", resp.json())

    def test_login_with_email(self):
        user = self.get_user()

        data = {"username": "hi@foo.com", "password": "hello world"}
        resp = self.client.post("/s/login", data=data)
        self.assertEqual(resp.status_code, 400)

        user.set_password('hello world')
        user.save()

        obj = UserEmail.objects.create(user=user, email="hi@foo.com")
        resp = self.client.post("/s/login", data=data)
        self.assertEqual(resp.status_code, 400)

        obj.primary = True
        obj.verified = True
        obj.save()

        resp = self.client.post("/s/login", data=data)
        self.assertEqual(resp.status_code, 200)



import typing as t
from abc import ABCMeta
from django.contrib.auth.models import User as AuthUser
from ..models import get_tenant_model, Tenant, Member


class AbstractToken(metaclass=ABCMeta):
    scope: str
    tenant_id: t.Any
    user: AuthUser

    @staticmethod
    def get_tenant(tenant_id) -> Tenant:
        return get_tenant_model().objects.get_from_cache_by_pk(tenant_id)

    def get_all_permissions(self, tenant_id) -> t.Optional[t.Set[str]]:
        try:
            member = Member.objects.get_by_natural_key(tenant_id, self.user.pk)
            if member.is_active:
                return member.get_all_permissions()
        except Member.DoesNotExist:
            return None

    def check_permissions(self, resource_permissions: t.List[str], tenant_id=None) -> bool:
        if tenant_id is None:
            tenant_id = self.tenant_id

        if not tenant_id or not self.user:
            return False

        if not self.user.is_active:
            return False

        if self.user.is_superuser:
            return True

        # tenant owner has full permission
        tenant = self.get_tenant(tenant_id)
        if tenant.owner_id == self.user.pk:
            return True

        perms = self.get_all_permissions(tenant_id)
        if not perms:
            return False

        for name in resource_permissions:
            if name in perms:
                return True
        return False

    def check_scopes(self, resource_scopes: t.List[str]):
        if self.scope == "__all__":
            return True

        token_scopes = set(self.scope.split())
        for rs in resource_scopes:
            if set(rs.split()).issubset(token_scopes):
                return True
        return False


class AuthToken(AbstractToken):
    scope = "__all__"

    def __init__(self, user=None, tenant_id=None):
        self.user = user
        self.tenant_id = tenant_id



from django.apps import AppConfig
from django.conf import settings
from django.db.models.signals import post_migrate
from ._management import create_permissions

if not hasattr(settings, 'SAAS_TENANT_MODEL'):
    setattr(settings, 'SAAS_TENANT_MODEL', 'saas_base.Tenant')


class CoreConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "saas_base"
    verbose_name = "SaaS"

    def ready(self):
        post_migrate.connect(create_permissions)



from django.contrib import admin
from .models import (
    Permission,
    Group,
    Member,
    UserEmail,
    get_tenant_model,
)


@admin.register(Permission)
class PermissionAdmin(admin.ModelAdmin):
    list_display = ["name", "internal", "created_at"]


@admin.register(get_tenant_model())
class TenantAdmin(admin.ModelAdmin):
    list_display = ["pk", "slug", "owner", "expires_at", "created_at"]


@admin.register(Group)
class GroupAdmin(admin.ModelAdmin):
    pass


@admin.register(Member)
class MemberAdmin(admin.ModelAdmin):
    list_display = ["pk", "name", "invite_email", "status", "created_at"]


@admin.register(UserEmail)
class UserEmailAdmin(admin.ModelAdmin):
    list_display = ["pk", "email", "primary", "verified", "created_at"]



import typing as t
import requests
from .base import BaseMailProvider


class MailgunProvider(BaseMailProvider):
    name = "mailgun"

    def get_api_url(self) -> str:
        domain = self.options["domain"]
        return f"https://api.mailgun.net/v3/{domain}/messages"

    def get_api_auth(self):
        return "api", self.options["api_key"]

    def send_mail(
            self,
            subject: str,
            recipients: t.List[str],
            text_message: str,
            html_message: t.Optional[str] = None,
            from_email: t.Optional[str] = None,
            headers: t.Optional[t.Dict[str, str]] = None,
            reply_to: t.Optional[str] = None,
            fail_silently: bool = False):
        if from_email is None:
            from_email = self.default_from_email
        url = self.get_api_url()
        auth = self.get_api_auth()
        data = {
            "from": from_email,
            "to": recipients,
            "subject": subject,
            "text": text_message,
            "html": html_message,
        }
        if reply_to:
            data["h:Reply-To"] = reply_to
        for k in headers:
            data[f"h:{k}"] = headers[k]
        return requests.post(url, auth=auth, data=data)



from .mail import get_mail_provider
from .settings import saas_settings
from .signals import before_send_mail


def send_mail(sender, subject: str, template_id: str, recipients, **context):
    context.setdefault('site', saas_settings.SITE)
    provider = get_mail_provider('notification', 'default')
    text_message, html_message = provider.render_message(template_id, context)
    before_send_mail.send(
        sender,
        subject=subject,
        recipients=recipients,
        text_message=text_message,
        html_message=html_message,
    )
    if saas_settings.MAIL_IMMEDIATE_SEND:
        return provider.send_mail(
            subject,
            recipients,
            text_message,
            html_message,
        )



from django.utils.translation import gettext_lazy as _
from django.conf import settings
from django.contrib.auth import login, logout
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.throttling import AnonRateThrottle
from ..models import Member
from ..drf.views import Endpoint
from ..settings import saas_settings
from ..security import check_security_rules
from ..serializers.auth import (
    EmailCode,
    SignupRequestCodeSerializer,
    SignupCreateUserSerializer,
    SignupConfirmCodeSerializer,
    SignupConfirmPasswordSerializer,
)
from ..serializers.password import PasswordLoginSerializer
from ..signals import after_signup_user, after_login_user
from .._notification import send_mail


class SignupRequestEndpoint(Endpoint):
    email_template_id = "signup_code"
    email_subject = _("Signup Request")
    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]

    def get_serializer_class(self):
        if saas_settings.SIGNUP_REQUEST_CREATE_USER:
            return SignupCreateUserSerializer
        return SignupRequestCodeSerializer

    def post(self, request: Request):
        """Send a sign-up code to user's email address."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj: EmailCode = serializer.save()

        # check bad request rules
        check_security_rules(saas_settings.SIGNUP_SECURITY_RULES, request)

        send_mail(
            self.__class__,
            self.email_subject,
            self.email_template_id,
            recipients=[obj.recipient()],
            code=obj.code,
        )
        return Response(status=204)


class AuthEndpoint(Endpoint):
    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]

    def login_user(self, request: Request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.save()
        login(request._request, user, 'django.contrib.auth.backends.ModelBackend')
        return user


class SignupConfirmEndpoint(AuthEndpoint):
    def get_serializer_class(self):
        if saas_settings.SIGNUP_REQUEST_CREATE_USER:
            return SignupConfirmCodeSerializer
        return SignupConfirmPasswordSerializer

    def post(self, request: Request):
        """Register a new user and login."""
        user = self.login_user(request)
        # update related membership
        Member.objects.filter(invite_email=user.email).update(
            user=user,
            status=Member.InviteStatus.WAITING
        )
        after_signup_user.send(
            self.__class__,
            user=user,
            request=request,
            strategy="password",
        )
        return Response({"next": settings.LOGIN_REDIRECT_URL})


class PasswordLogInEndpoint(AuthEndpoint):
    serializer_class = PasswordLoginSerializer

    def post(self, request: Request):
        """Login a user with the given username and password."""
        user = self.login_user(request)
        after_login_user.send(
            self.__class__,
            user=user,
            request=request,
            strategy="password",
        )
        return Response({"next": settings.LOGIN_REDIRECT_URL})


class LogoutEndpoint(AuthEndpoint):
    def post(self, request: Request):
        """Clear the user session and log the user out."""
        logout(request._request)
        return Response({"next": settings.LOGIN_URL})



from rest_framework.permissions import BasePermission
from rest_framework.request import Request
from ..settings import saas_settings

__all__ = [
    'HasResourcePermission',
    'HasResourceScope',
]


http_method_actions = {
    'GET': 'read',
    'HEAD': 'read',
    'POST': 'write',
    'PUT': 'write',
    'PATCH': 'write',
    'DELETE': 'admin',
}


class HasResourcePermission(BasePermission):
    @staticmethod
    def get_resource_permissions(view, method):
        resource = getattr(view, 'resource_name', None)
        if not resource:
            return

        action = getattr(view, 'resource_action', None)
        if not action:
            method_actions = getattr(view, 'resource_http_method_actions', http_method_actions)
            action = method_actions.get(method)

        permission = saas_settings.PERMISSION_NAME_FORMATTER.format(
            resource=resource,
            action=action,
        )
        return [permission]

    def has_permission(self, request: Request, view):
        if not request.user or not request.user.is_active:
            return False
        resource_permissions = self.get_resource_permissions(view, request.method)
        if not resource_permissions:
            return True
        if request.auth and hasattr(request.auth, 'check_permissions'):
            return request.auth.check_permissions(resource_permissions)
        return False


class HasResourceScope(BasePermission):
    @staticmethod
    def get_resource_scopes(view, method):
        if hasattr(view, 'get_resource_scopes'):
            resource_scopes = view.get_resource_scopes(method)
        elif hasattr(view, 'resource_scopes'):
            resource_scopes = view.resource_scopes
        else:
            resource_scopes = None
        return resource_scopes

    def has_permission(self, request: Request, view):
        resource_scopes = self.get_resource_scopes(view, request.method)
        if not resource_scopes:
            return True

        if request.auth and hasattr(request.auth, 'check_scopes'):
            return request.auth.check_scopes(resource_scopes)
        return False



import typing as t
import css_inline
from django.template import loader
from abc import ABCMeta, abstractmethod


class BaseMailProvider(metaclass=ABCMeta):
    name: str = "base"

    def __init__(self, **options):
        self.default_from_email = options.pop('default_from_email', None)
        self.options = options

    @abstractmethod
    def send_mail(
            self,
            subject: str,
            recipients: t.List[str],
            text_message: str,
            html_message: t.Optional[str] = None,
            from_email: t.Optional[str] = None,
            headers: t.Optional[t.Dict[str, str]] = None,
            reply_to: t.Optional[str] = None,
            fail_silently: bool = False):
        pass

    def render_message(
            self,
            template_id: str,
            context: t.Dict[str, t.Any],
            using: t.Optional[str] = None) -> t.Tuple[str, str]:
        text: str = loader.render_to_string(
            f"saas_emails/{template_id}.text",
            context, using=using
        )
        html: str = loader.render_to_string(
            f"saas_emails/{template_id}.html",
            context, using=using
        )
        return text, css_inline.inline(html)



from django.urls import path
from .password import (
    PasswordResetEndpoint,
    PasswordForgotEndpoint,
)
from .auth import (
    LogoutEndpoint,
    PasswordLogInEndpoint,
    SignupRequestEndpoint,
    SignupConfirmEndpoint,
)


urlpatterns = [
    path('logout', LogoutEndpoint.as_view()),
    path('login', PasswordLogInEndpoint.as_view()),
    path('signup/request', SignupRequestEndpoint.as_view()),
    path('signup/confirm', SignupConfirmEndpoint.as_view()),
    path('password/forgot', PasswordForgotEndpoint.as_view()),
    path('password/reset', PasswordResetEndpoint.as_view()),
]



from rest_framework.request import Request
from rest_framework.authentication import (
    SessionAuthentication as _SessionAuthentication,
)
from .token import AuthToken


class SessionAuthentication(_SessionAuthentication):
    def authenticate(self, request: Request):
        tenant_id = getattr(request._request, "tenant_id", None)
        credentials = super().authenticate(request)
        if credentials is None:
            return None
        user = credentials[0]
        token = AuthToken(user, tenant_id)
        return user, token



from rest_framework.request import Request


def get_client_ip(request: Request):
    ip = request.headers.get('CF-Connecting-IP')
    if ip:
        return ip

    x_forwarded_for = request.headers.get('X-Forwarded-For')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
        return ip

    ip = request.headers.get('X-Real-IP')
    if ip:
        return ip
    return request.META.get('REMOTE_ADDR')



import re
from django.core import mail
from tests.client import FixturesTestCase
from saas_base.models import UserEmail


class TestPasswordAPI(FixturesTestCase):
    user_id = FixturesTestCase.OWNER_USER_ID

    def setUp(self):
        user = self.get_user()
        UserEmail.objects.create(user=user, email="hi@foo.com", primary=True, verified=True)

    def test_reset_password(self):
        data = {"email": "hi@foo.com"}
        resp = self.client.post("/s/password/forgot", data=data)
        self.assertEqual(resp.status_code, 204)
        self.assertEqual(len(mail.outbox), 1)
        msg = mail.outbox[0]
        codes = re.findall(r'Code: (\w{6})', msg.body)

        data = {**data, "code": codes[0], "password": "this is me"}
        resp = self.client.post("/s/password/reset", data=data)
        self.assertEqual(resp.status_code, 200)
        self.assertIn("next", resp.json())

    def test_wrong_email(self):
        data = {"email": "404@foo.com"}
        resp = self.client.post("/s/password/forgot", data=data)
        self.assertEqual(resp.status_code, 400)

    def test_wrong_code(self):
        data = {"email": "hi@foo.com"}
        resp = self.client.post("/s/password/forgot", data=data)
        self.assertEqual(resp.status_code, 204)
        data = {**data, "code": "AAAAAA", "password": "this is me"}
        resp = self.client.post("/s/password/reset", data=data)
        self.assertEqual(resp.status_code, 400)



from django.urls import path
from .tenant import (
    SelectedTenantEndpoint,
    CurrentMemberEndpoint,
    TenantsEndpoint,
)
from .members import (
    MemberListEndpoint,
    MemberItemEndpoint,
    MemberGroupsEndpoint,
    MemberGroupItemEndpoint,
    MemberPermissionsEndpoint,
    MemberPermissionItemEndpoint,
)
from .user import (
    UserEndpoint,
    UserPasswordEndpoint,
    UserEmailListEndpoint,
    UserTenantsEndpoint,
)

urlpatterns = [
    path('user', UserEndpoint.as_view()),
    path('user/password', UserPasswordEndpoint.as_view()),
    path('user/emails', UserEmailListEndpoint.as_view()),
    path('user/tenants', UserTenantsEndpoint.as_view()),

    path('tenants', TenantsEndpoint.as_view()),
    path('tenant', SelectedTenantEndpoint.as_view()),
    path('tenant/member', CurrentMemberEndpoint.as_view()),

    path('members', MemberListEndpoint.as_view()),
    path('members/<pk>', MemberItemEndpoint.as_view()),
    path('members/<member_id>/groups', MemberGroupsEndpoint.as_view()),
    path('members/<member_id>/permissions', MemberPermissionsEndpoint.as_view()),
    path('members/<member_id>/groups/<pk>', MemberGroupItemEndpoint.as_view()),
    path('members/<member_id>/permissions/<pk>', MemberPermissionItemEndpoint.as_view()),
]



import typing as t
from .base import BaseMailProvider
from .django import DjangoMailProvider
from ..settings import saas_settings


def get_mail_provider(name: str = 'default', fallback: t.Optional[str] = None):
    if name in saas_settings.MAIL_PROVIDERS:
        return saas_settings.MAIL_PROVIDERS[name]
    if fallback:
        return saas_settings.MAIL_PROVIDERS[fallback]


__all__ = ['get_mail_provider', 'BaseMailProvider', 'DjangoMailProvider']
